from anndata import read_h5ad
import sys
from time import time
from scipy import stats, sparse
import numpy as np
import collections
import pickle
from sklearn.preprocessing import normalize
import os
from collections import Counter
from scipy import spatial
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score,accuracy_score,precision_recall_fscore_support, cohen_kappa_score, auc, average_precision_score,f1_score,precision_recall_curve
import time
import copy
from sklearn import preprocessing
from sklearn.metrics.pairwise import cosine_similarity
#from libs import *
from sklearn.utils.graph_shortest_path import graph_shortest_path
from scipy.sparse.linalg import svds, eigs

def get_ontology_parents(GO_net, g, dfs_depth):
	term_valid = set()
	ngh_GO = set()
	ngh_GO.add(g)
	depth = {}
	depth[g] = 0
	while len(ngh_GO) > 0:
		for GO in list(ngh_GO):
			for GO1 in GO_net[GO]:
				ngh_GO.add(GO1)
				depth[GO1] = depth[GO] + 1
			ngh_GO.remove(GO)
			if depth[GO] < dfs_depth:
				term_valid.add(GO)
	return term_valid

def creat_cell_ontology_matrix(train_Y, co2co_graph, cell_ontology_ids, dfs_depth):
	lset = set(cell_ontology_ids)

	seen_l = sorted(np.unique(train_Y))
	unseen_l = sorted(lset - set(train_Y))
	ys =  np.concatenate((seen_l, unseen_l))
	i2l = {}
	l2i = {}
	for l in ys:
		nl = len(i2l)
		l2i[l] = nl
		i2l[nl] = l
	nco = len(i2l)

	net_dict = collections.defaultdict(dict)
	net_mat = np.zeros((nco,nco))
	for co1 in co2co_graph:
		l1 = l2i[co1]
		for co2 in co2co_graph[co1]:
			l2 = l2i[co2]
			net_dict[l1][l2] = 1
			net_mat[l1][l2] = 1
	for n in range(nco):
		ngh = get_ontology_parents(net_dict, n, dfs_depth)
		net_dict[n][n] = 1
		for n1 in ngh:
			net_dict[n][n1] = 1
	return unseen_l, l2i, i2l, net_dict, net_mat


def create_propagate_networks_using_nlp(l2i, onto_net, cls2cls, co2vec_nlp, rsts = [0.5,0.6,0.7,0.8], diss=[2,3], thress=[1,0.8]):
	ncls = np.shape(cls2cls)[0]
	#onto_net_nlp, onto_net_bin, stack_net_nlp, stack_net_bin, onto_net_nlp_all_pairs = create_nlp_networks(l2i, onto_net, cls2cls, ontology_nlp_file)
	#ncls = np.shape(cls2cls)[0]
	onto_net_nlp_all_pairs = (cosine_similarity(co2vec_nlp) + 1 ) /2#1 - spatial.distance.cosine(onto_nlp_emb, onto_nlp_emb)
	onto_net_nlp = np.zeros((ncls, ncls))
	stack_net_nlp = np.zeros((ncls, ncls))

	for n1 in onto_net:
		for n2 in onto_net[n1]:
			if n1==n2:
				continue
			stack_net_nlp[n2,n1] = onto_net_nlp_all_pairs[n2, n1]
			stack_net_nlp[n1,n2] = onto_net_nlp_all_pairs[n1, n2]
	for n1 in range(ncls):
		for n2 in range(ncls):
			if cls2cls[n1,n2] == 1 or cls2cls[n2,n1] == 1:
				onto_net_nlp[n1,n2] = onto_net_nlp_all_pairs[n1, n2]
				onto_net_nlp[n2,n1] = onto_net_nlp_all_pairs[n2, n1]
	#network = create_consensus_networks(rsts, stack_net_nlp, onto_net_nlp_all_pairs, cls2cls)
	cls2cls_sp = graph_shortest_path(cls2cls,method='FW',directed =False)
	networks = []
	for rst in rsts:
		for dis in diss:
			for thres in thress:
				#use_net = np.copy(stack_net_nlp)
				use_net = np.copy(onto_net_nlp)
				use_net[(cls2cls_sp<=dis)&(onto_net_nlp_all_pairs > thres)] = onto_net_nlp_all_pairs[(cls2cls_sp<=dis)&(onto_net_nlp_all_pairs > thres)]
				onto_net_rwr = RandomWalkRestart(use_net, rst)
				networks.append(onto_net_rwr)
	return networks


def map_genes(test_X, test_genes, train_genes):
	ntest_cell = np.shape(test_X)[0]
	ntrain_gene = len(train_genes)
	new_test_x = np.zeros((ntest_cell, ntrain_gene))
	genes = set(test_genes) & set(train_genes)
	train_genes = list(train_genes)
	test_genes = list(test_genes)
	ind1 = []
	ind2 = []
	for i,g in enumerate(genes):
		ind1.append(train_genes.index(g))
		ind2.append(test_genes.index(g))
	ind1 = np.array(ind1)
	ind2 = np.array(ind2)
	new_test_x[:,ind1] = test_X[:,ind2]
	return new_test_x


def extend_prediction_2unseen(pred_Y_seen, networks, nseen, ratio=200, use_normalize=False):
	if not isinstance(networks, list):
		networks = [networks]
	pred_Y_all_totoal = 0.
	for onto_net_rwr in networks:
		if use_normalize:
			onto_net_rwr = onto_net_rwr - np.tile(np.mean(onto_net_rwr, axis = 1), (np.shape(onto_net_rwr)[0], 1))
		pred_Y_seen_norm = pred_Y_seen / pred_Y_seen.sum(axis=1)[:, np.newaxis]
		pred_Y_all = np.dot(pred_Y_seen_norm, onto_net_rwr[:nseen,:])
		pred_Y_all[:,:nseen] = normalize(pred_Y_all[:,:nseen],norm='l1',axis=1)
		pred_Y_all[:,nseen:] = normalize(pred_Y_all[:,nseen:],norm='l1',axis=1) * ratio
		pred_Y_all_totoal += pred_Y_all
	return pred_Y_all_totoal

def create_consensus_networks(rsts, onto_net_mat, onto_net_nlp_all_pairs, cls2cls, diss=[2,3], thress=[1,0.8]):
	cls2cls_sp = graph_shortest_path(cls2cls,method='FW',directed =False)
	ncls = np.shape(onto_net_mat)[0]
	networks = []
	for rst in rsts:
		for dis in diss:
			for thres in thress:
				use_net = np.copy(onto_net_mat)
				use_net[(cls2cls_sp<=dis)&(onto_net_nlp_all_pairs > thres)] = onto_net_nlp_all_pairs[(cls2cls_sp<=dis)&(onto_net_nlp_all_pairs > thres)]
				onto_net_rwr = RandomWalkRestart(use_net, rst)
				networks.append(onto_net_rwr)
	return networks


def fine_nearest_co_using_nlp(sentences,co2emb,cutoff=0.8):
	from sentence_transformers import SentenceTransformer
	model = SentenceTransformer('bert-base-nli-mean-tokens')
	sentence_embeddings = model.encode(sentences)
	co_embeddings = []
	cos = []
	for co in co2emb:
		co_embeddings.append(co2emb[co])
		cos.append(co)
	co_embeddings = np.array(co_embeddings)
	sent2co = {}
	for sentence, embedding, ind in zip(sentences, sentence_embeddings, range(len(sentences))):
		scs = cosine_similarity(co_embeddings, embedding.reshape(1,-1))

		co_id = np.argmax(scs)
		sc = scs[co_id]
		if sc>cutoff:
			sent2co[sentence] = cos[co_id]
	return sent2co


def read_cell_type_nlp_network(nlp_emb_file, cell_type_network_file):
	cell_ontology_ids = set()
	fin = open(cell_type_network_file)
	co2co_graph = {}
	for line in fin:
		w = line.strip().split('\t')
		if w[0] not in co2co_graph:
			co2co_graph[w[0]] = set()
		co2co_graph[w[0]].add(w[1])
		cell_ontology_ids.add(w[0])
		cell_ontology_ids.add(w[1])
	fin.close()
	if nlp_emb_file is not None:
		fin = open(nlp_emb_file)
		co2vec_nlp = {}
		for line in fin:
			w = line.strip().split('\t')
			vec = []
			for i in range(1,len(w)):
				vec.append(float(w[i]))
			co2vec_nlp[w[0]] = np.array(vec)
		fin.close()
		co2co_nlp = {}
		for id1 in co2co_graph:
			co2co_nlp[id1] = {}
			for id2 in co2co_graph[id1]:
				sc = 1 - spatial.distance.cosine(co2vec_nlp[id1], co2vec_nlp[id2])
				co2co_nlp[id1][id2] = sc
	else:
		co2co_nlp = {}
		for id1 in co2co_graph:
			co2co_nlp[id1] = {}
			for id2 in co2co_graph[id1]:
				co2co_nlp[id1][id2] = 1.
		co2vec_nlp = {}
		for c in cell_ontology_ids:
			co2vec_nlp[c] = np.ones((10))
	return co2co_graph, co2co_nlp, co2vec_nlp, cell_ontology_ids

def create_unseen_candidates(cell_type_network_file, co2i, i2co, nseen, use_unseen_distance, test_Y_pred_all):
	nct = len(co2i)
	A = np.zeros((nct, nct))
	fin = open(cell_type_network_file)
	for line in fin:
		w = line.strip().split('\t')
		A[co2i[w[0]], co2i[w[1]]] = 1.
		A[co2i[w[1]], co2i[w[0]]] = 1.
	fin.close()
	A_dis = graph_shortest_path(A,method='FW',directed =False)
	min_d = np.min(A_dis[:nseen, nseen:], axis = 0)
	assert(len(min_d) == nct - nseen)
	unseen_cand = np.where(min_d > use_unseen_distance)[0] + nseen
	test_Y_pred_all[:, unseen_cand] = 0
	assert(np.shape(test_Y_pred_all)[1] == nct)
	return test_Y_pred_all



def graph_embedding_dca(A, i2l, mi=0, dim=20,unseen_l=None):
	nl = np.shape(A)[0]
	seen_ind = []
	unseen_ind = []
	for i in range(nl):
		if i2l[i] in unseen_l:
			unseen_ind.append(i)
		else:
			seen_ind.append(i)
	seen_ind = np.array(seen_ind)
	unseen_ind = np.array(unseen_ind)

	#if len(seen_ind) * 0.8 < dim:
	#	dim = int(len(seen_ind) * 0.8)
	if mi==0 or mi == 1:
		sp = graph_shortest_path(A,method='FW',directed =False)
	else:
		sp = RandomWalkRestart(A, 0.8)

	sp = sp[seen_ind, :]
	sp = sp[:,seen_ind]
	X = np.zeros((np.shape(sp)[0],dim))
	svd_dim = min(dim, np.shape(sp)[0]-1)
	if mi==0 or mi == 2:
		X[:,:svd_dim] = svd_emb(sp, dim=svd_dim)
	else:
		X[:,:svd_dim] = DCA_vector(sp, dim=svd_dim)[0]
	X_ret = np.zeros((nl, dim))
	X_ret[seen_ind,:] = X
	if mi==2 or mi == 3:
		sp *= -1
	return sp, X_ret

def DCA_vector(Q, dim):
	nnode = Q.shape[0]
	alpha = 1. / (nnode **2)
	Q = np.log(Q + alpha) - np.log(alpha);

	#Q = Q * Q';
	[U, S, V] = svds(Q, dim);
	S = np.diag(S)
	X = np.dot(U, np.sqrt(S))
	Y = np.dot(np.sqrt(S), V)
	Y = np.transpose(Y)
	return X,U,S,V,Y

def RandomWalkRestart(A, rst_prob, delta = 1e-4, reset=None, max_iter=50,use_torch=False,return_torch=False):
	if use_torch:
		device = torch.device("cuda:0")
	nnode = A.shape[0]
	#print nnode
	if reset is None:
		reset = np.eye(nnode)
	nsample,nnode = reset.shape
	#print nsample,nnode
	P = renorm(A)
	P = P.T
	norm_reset = renorm(reset.T)
	norm_reset = norm_reset.T
	if use_torch:
		norm_reset = torch.from_numpy(norm_reset).float().to(device)
		P = torch.from_numpy(P).float().to(device)
	Q = norm_reset

	for i in range(1,max_iter):
		#Q = gnp.garray(Q)
		#P = gnp.garray(P)
		if use_torch:
			Q_new = rst_prob*norm_reset + (1-rst_prob) * torch.mm(Q, P)#.as_numpy_array()
			delta = torch.norm(Q-Q_new, 2)
		else:
			Q_new = rst_prob*norm_reset + (1-rst_prob) * np.dot(Q, P)#.as_numpy_array()
			delta = np.linalg.norm(Q-Q_new, 'fro')
		Q = Q_new
		#print (i,Q)
		sys.stdout.flush()
		if delta < 1e-4:
			break
	if use_torch and not return_torch:
		Q = Q.cpu().numpy()
	return Q


def renorm(X):
	Y = X.copy()
	Y = Y.astype(float)
	ngene,nsample = Y.shape
	s = np.sum(Y, axis=0)
	#print s.shape()
	for i in range(nsample):
		if s[i]==0:
			s[i] = 1
			if i < ngene:
				Y[i,i] = 1
			else:
				for j in range(ngene):
					Y[j,i] = 1. / ngene
		Y[:,i] = Y[:,i]/s[i]
	return Y

def mean_normalization(train_X_mean, test_X):
	test_X = np.log1p(test_X)
	test_X_mean = np.mean(test_X, axis = 0)
	test_X = test_X - test_X_mean + train_X_mean
	return test_X

def process_expression(train_X, test_X, train_genes, test_genes):
	#this data process function is adapted from ACTINN, please check ACTINN for more information.
	#test_X = map_genes(test_X, test_genes, train_genes)
	c2g = np.vstack([train_X, test_X])
	c2g = np.array(c2g,  dtype=np.float64)
	c2g = c2g.T
	index = np.sum(c2g, axis=1)>0
	c2g = c2g[index, :]
	train_genes = train_genes[index]
	c2g = np.divide(c2g, np.sum(c2g, axis=0, keepdims=True)) * 10000
	c2g = np.log2(c2g+1)
	expr = np.sum(c2g, axis=1)
	#total_set = total_set[np.logical_and(expr >= np.percentile(expr, 1), expr <= np.percentile(expr, 99)),]
	index = np.logical_and(expr >= np.percentile(expr, 1), expr <= np.percentile(expr, 99))
	c2g = c2g[index,]
	train_genes = train_genes[index]
	#print (np.logical_and(expr >= np.percentile(expr, 1), expr <= np.percentile(expr, 99)))

	cv = np.std(c2g, axis=1) / np.mean(c2g, axis=1)
	index = np.logical_and(cv >= np.percentile(cv, 1), cv <= np.percentile(cv, 99))
	c2g = c2g[index,]
	train_genes = train_genes[index]
	c2g = c2g.T
	c2g_list_new = []
	index = 0
	for c in [train_X, test_X]:
		ncell = np.shape(c)[0]
		c2g_list_new.append(c2g[index:index+ncell,:])
		index = ncell
		assert (len(train_genes) == np.shape(c2g)[1])
	return c2g_list_new[0], c2g_list_new[1], train_genes

def emb_ontology(i2l, ontology_mat, co2co_nlp, dim=5, mi=0, unseen_l = None):
	nco = len(i2l)
	network = np.zeros((nco, nco))
	for i in range(nco):
		c1 = i2l[i]
		for j in range(nco):
			if ontology_mat[i,j] == 1:
				network[i,j] = co2co_nlp[c1][i2l[j]]
				network[j,i] = co2co_nlp[c1][i2l[j]]
	idd = 0
	sp, i2emb = graph_embedding_dca(network, i2l, mi=mi, dim=dim, unseen_l=unseen_l)
	return i2emb



import numpy as np
#import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
tf.logging.set_verbosity(tf.logging.ERROR)
import sys
from scipy.special import softmax
import time
import math
from sklearn.metrics import roc_auc_score, average_precision_score
from scipy import stats

class BilinearNN:
	def __init__(self, Y_emb, nseen, ngene, use_pretrain = None, nhidden=[1000], l2=0.005):
		tf.reset_default_graph()
		self.ncls, self.ndim = np.shape(Y_emb)
		self.l2 = l2
		self.nseen = nseen
		self.ngene = ngene

		self.use_pretrain = use_pretrain
		self.seen_Y_emb = Y_emb[:nseen,:]
		self.nY = np.shape(self.seen_Y_emb)[0]
		self.seen_Y_emb = np.array(self.seen_Y_emb, dtype=np.float32)
		self.Y_emb = np.array(Y_emb, dtype=np.float32)

		self.nhidden = [self.ngene]
		self.nhidden.extend(nhidden)
		self.nhidden.append(self.ndim)
		tf.set_random_seed(3)

		self.__build()
		self.__build_loss()
		self.saver = tf.train.Saver()
		self.sess = tf.Session(config=tf.ConfigProto(gpu_options=tf.GPUOptions(allow_growth=True)))
		if self.use_pretrain is not None:
			self.saver.restore(self.sess, self.use_pretrain)
		#else:
		#	self.sess.run(tf.global_variables_initializer())


	def read_training_data(self, train_X, train_Y, use_valid = False, valid_X = None, valid_Y = None, test_X = None, test_Y = None):
		self.use_valid = use_valid
		if self.use_valid:
			if valid_X is None:
				np.random.seed(1)
				nx = np.shape(train_X)[0]
				ntrain = int(nx*0.9)
				permutation = list(np.random.permutation(nx))
				train_ind = permutation[:ntrain]
				valid_ind = permutation[ntrain:]
				self.train_X = train_X[train_ind, :]
				self.valid_X = train_X[valid_ind, :]
				self.train_Y = train_Y[train_ind]
				self.valid_Y = train_Y[valid_ind]
			else:
				self.train_X = train_X
				self.valid_X = valid_X
				self.train_Y = train_Y
				self.valid_Y = valid_Y
			self.valid_Y = self.one_hot_matrix(self.valid_Y, self.ncls)
		else:
			self.train_X = train_X
			self.train_Y = train_Y
		self.train_Y = self.one_hot_matrix(self.train_Y, self.nseen)
		self.nX = np.shape(self.train_X)[0]

		if test_X is not None:
			self.test_X = test_X
			self.test_Y = self.one_hot_matrix(test_Y, self.ncls)
		else:
			self.test_X = None
			self.test_Y = None
		self.train_X = np.array(self.train_X, dtype=np.float32)
		self.train_Y = np.array(self.train_Y, dtype=np.float32)


	def __build(self,stddev=0.0001, seed=3):#W,H,B
		tf.set_random_seed(seed) # set seed to make the results consistant
		#w_init =  tf.contrib.layers.xavier_initializer(seed = seed)
		w_init =  tf.glorot_uniform_initializer(seed = seed)#tf 2.0 has no tf.contribu. Glorot is the same as xavier
		b_init = tf.zeros_initializer()
		self.nlayer = len(self.nhidden)
		self.W = {}
		self.B = {}
		for i in range(1,self.nlayer):
			self.W[i] = tf.get_variable("W"+str(i), [self.nhidden[i], self.nhidden[i-1]], initializer = w_init, dtype=tf.float32)
			self.B[i] = tf.get_variable("B"+str(i), [self.nhidden[i], 1], initializer = b_init, dtype=tf.float32)
		#self.B = tf.get_variable("B", [1, 1], initializer = tf.zeros_initializer(), dtype=tf.float32)

	def __build_loss(self):
		self.mini_train_X = tf.placeholder(shape=[None, self.ngene], dtype=tf.float32)
		self.mini_train_Y = tf.placeholder(shape=[None, self.nseen], dtype=tf.float32)
		self.mini_train_Y_emb = tf.placeholder(shape=[None, None], dtype=tf.float32)
		self.train_keep_prob = tf.placeholder(tf.float32)
		self.p = self.mini_train_X
		self.loss = 0
		for i in range(1,self.nlayer):
			self.p = tf.add(tf.matmul(self.p, tf.transpose(self.W[i])), tf.transpose(self.B[i]))
			if i != self.nlayer - 1:
				self.p = tf.nn.relu(self.p)
				self.p = tf.nn.dropout(self.p, self.train_keep_prob)
			self.loss += (tf.nn.l2_loss(self.W[i])) * self.l2
		self.p = tf.matmul(self.p, tf.transpose(self.mini_train_Y_emb))
		self.label = self.mini_train_Y
		self.entropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits = self.p, labels = self.label)
		self.loss += tf.reduce_mean(self.entropy)

	# Turn labels into matrix
	def one_hot_matrix(self, labels, C):
		C = tf.constant(C, name = "C")
		one_hot_matrix = tf.one_hot(labels, C, axis = 0)
		sess_tmp = tf.Session()
		one_hot = sess_tmp.run(one_hot_matrix)
		sess_tmp.close()
		return one_hot.T

	def predict(self, test_X):
		test_X = np.array(test_X, dtype=np.float32)
		p = self.sess.run(self.p, feed_dict={self.mini_train_X:test_X, self.mini_train_Y_emb: self.seen_Y_emb, self.train_keep_prob: 1.0})
		test_Y_pred = softmax(p, axis=1)
		return test_Y_pred

	def predict_prob(self,  X, Y, Y_emb):
		label = Y
		p = self.sess.run(self.p, feed_dict={self.mini_train_X: X, self.mini_train_Y_emb: Y_emb, self.train_keep_prob: 1.0})
		p = softmax(p, axis=1)
		accuracy = np.mean(np.argmax(p,axis=1) == np.argmax(label,axis=1))
		[nsample, nclass] = np.shape(Y)
		class_auc_macro = np.full(nclass, np.nan)
		class_auprc_macro =  np.full(nclass, np.nan)
		for i in range(self.nseen):
			if len(np.unique(Y[:,i]))==2:
				class_auc_macro[i] = roc_auc_score(Y[:,i], p[:,i])
				#auc_tmp = roc_auc_score(Y[:,i], p[:,i]*1e50)
				class_auprc_macro[i] = average_precision_score(Y[:,i], p[:,i])
				#auprc_tmp = average_precision_score(Y[:,i], p[:,i]*1e50)
				#print (class_auprc_macro[i],  class_auc_macro[i])
		seen_auroc = np.nanmedian(class_auc_macro)
		seen_auprc = np.nanmedian(class_auprc_macro)
		class_auc_macro = np.full(nclass, np.nan)
		class_auprc_macro =  np.full(nclass, np.nan)
		for i in range(self.nseen, self.ncls):
			if i >= np.shape(Y)[1] or i  >= np.shape(p)[1]:
				break
			if len(np.unique(Y[:,i]))==2:
				class_auc_macro[i] = roc_auc_score(Y[:,i], p[:,i])
				class_auprc_macro[i] = average_precision_score(Y[:,i], p[:,i])
		unseen_auroc = np.nanmedian(class_auc_macro)
		unseen_auprc = np.nanmedian(class_auprc_macro)

		return accuracy, p, seen_auroc, seen_auprc, unseen_auroc, unseen_auprc

	# Get the mini batches
	def random_mini_batches(self, X, Y, mini_batch_size=32, seed=1):
		# input -- X (training set), Y (true labels)
		# output -- mini batches
		ns = X.shape[0]
		mini_batches = []
		np.random.seed(seed)
		# shuffle (X, Y)
		permutation = list(np.random.permutation(ns))
		shuffled_X = X[permutation, :]
		shuffled_Y = Y[permutation, :]
		# partition (shuffled_X, shuffled_Y), minus the end case.
		num_complete_minibatches = int(math.floor(ns/mini_batch_size)) # number of mini batches of size mini_batch_size in your partitionning
		for k in range(0, num_complete_minibatches):
			mini_batch_X = shuffled_X[k * mini_batch_size : k * mini_batch_size + mini_batch_size,:]
			mini_batch_Y = shuffled_Y[k * mini_batch_size : k * mini_batch_size + mini_batch_size,:]
			mini_batch = (mini_batch_X, mini_batch_Y)
			mini_batches.append(mini_batch)
		# handling the end case (last mini-batch < mini_batch_size)
		if ns % mini_batch_size != 0:
			mini_batch_X = shuffled_X[num_complete_minibatches * mini_batch_size : ns,:]
			mini_batch_Y = shuffled_Y[num_complete_minibatches * mini_batch_size : ns,:]
			mini_batch = (mini_batch_X, mini_batch_Y)
			mini_batches.append(mini_batch)
		return mini_batches

	def optimize(self, max_iter = 50, minibatch_size = 128, keep_prob = 1., lr = 0.0001, save_model = None):

		self.keep_prob = keep_prob
		tf.set_random_seed(3)
		seed = 3
		global_step = tf.Variable(0, trainable=False)
		decay_lr = tf.train.exponential_decay(lr, global_step, 1000, 0.95, staircase=True)
		train_op =  tf.train.AdamOptimizer(learning_rate=decay_lr).minimize(self.loss)
		#print (self.sess.run(self.W[1]))
		self.sess.run(tf.global_variables_initializer())
		for epoch in range(max_iter):
			seed = seed + 1
			minibatches = self.random_mini_batches(self.train_X, self.train_Y, minibatch_size, seed)
			epoch_cost = 0.
			num_minibatches = int(self.nX / minibatch_size)
			for minibatch in minibatches:
				(minibatch_X, minibatch_Y) = minibatch
				minibatch_cost, p, label, entropy,  _ = self.sess.run([self.loss, self.p, self.label, self.entropy, train_op], feed_dict={self.train_keep_prob: self.keep_prob,self.mini_train_X: minibatch_X, self.mini_train_Y: minibatch_Y, self.mini_train_Y_emb: self.seen_Y_emb})
				epoch_cost += minibatch_cost/ num_minibatches

			if (epoch+1) % 1 == 0:
				train_acc,_,train_auroc, train_auproc,_,_ = self.predict_prob(self.train_X, self.train_Y, self.seen_Y_emb)
				print ("Training cost after epoch %i: loss:%.6f acc: %.3f auc: %.3f auprc: %.3f" % (epoch+1, epoch_cost,train_acc,train_auroc,train_auproc))
				sys.stdout.flush()
				if self.use_valid:
					valid_acc,_,valid_auroc,valid_auprc, valid_unseen_auroc, valid_unseen_auprc = self.predict_prob(self.valid_X, self.valid_Y, self.Y_emb)
					print ("Validation cost after epoch %i: acc: %.3f auc: %.3f auprc: %.3f unseen_auc : %.3f unseen_auprc: %.3f" % (epoch+1, valid_acc,valid_auroc,valid_auprc, valid_unseen_auroc, valid_unseen_auprc))
				if self.test_X is not None:
					test_acc,_,test_auroc,test_auprc, test_unseen_auroc, test_unseen_auprc = self.predict_prob(self.test_X, self.test_Y, self.Y_emb)
					print ("Test cost after epoch %i: acc: %.3f auc: %.3f auprc: %.3f unseen_auc : %.3f unseen_auprc: %.3f" % (epoch+1, test_acc,test_auroc,test_auprc,test_unseen_auroc,test_unseen_auprc))
		if save_model is not None:
			self.saver.save(self.sess, save_model)



import numpy as np
import os
import sys
from sklearn.preprocessing import normalize

import warnings
from OnClass.OnClass_utils import *
from OnClass.BilinearNN import BilinearNN

class OnClassModel:
	def __init__(self, cell_type_network_file='../../OnClass_data/cell_ontology/cl.ontology', cell_type_nlp_emb_file='../../OnClass_data/cell_ontology/cl.ontology.nlp.emb'):
		self.cell_type_nlp_emb_file = cell_type_nlp_emb_file
		self.cell_type_network_file = cell_type_network_file
		self.co2co_graph, self.co2co_nlp, self.co2vec_nlp, self.cell_ontology_ids = read_cell_type_nlp_network(self.cell_type_nlp_emb_file, self.cell_type_network_file)
		#self.cell_type_nlp_network_file = self.cell_type_nlp_emb_file+'network'
		#fout = open(self.cell_type_nlp_network_file,'w')
		#for w1 in self.co2co_nlp:
		#	for w2 in self.co2co_nlp[w1]:
		#		fout.write(w1+'\t'+w2+'\t'+str(self.co2co_nlp[w1][w2])+'\n')
		#fout.close()

	def EmbedCellTypes(self, train_Y_str, dim=5, emb_method=3, use_pretrain = None, write2file=None):
		"""
		Embed the cell ontology
		Parameters
		----------
		cell_type_network_file : each line should be cell_type_1\tcell_type_2\tscore for weighted network or cell_type_1\tcell_type_2 for unweighted network
		dim: `int`, optional (500)
			Dimension of the cell type embeddings
		emb_method: `int`, optional (3)
			dimensionality reduction method
		use_pretrain: `string`, optional (None)
			use pretrain file. This should be the numpy file of cell type embeddings. It can read the one set in write2file parameter.
		write2file: `string`, optional (None)
			write the cell type embeddings to this file path
		Returns
		-------
		co2emb, co2i, i2co
			returns three dicts, cell type name to embeddings, cell type name to cell type id and cell type id to embeddings.
		"""

		self.unseen_co, self.co2i, self.i2co, self.ontology_dict, self.ontology_mat = creat_cell_ontology_matrix(train_Y_str, self.co2co_graph, self.cell_ontology_ids, dfs_depth = 3)
		self.nco = len(self.i2co)
		Y_emb = emb_ontology(self.i2co, self.ontology_mat, dim = dim, mi=emb_method, co2co_nlp = self.co2co_nlp, unseen_l = self.unseen_co)
		self.co2emb = np.column_stack((np.eye(self.nco), Y_emb))
		self.nunseen = len(self.unseen_co)
		self.nseen = self.nco - self.nunseen
		self.co2vec_nlp_mat = np.zeros((self.nco, len(self.co2vec_nlp[self.i2co[0]])))
		for i in range(self.nco):
			self.co2vec_nlp_mat[i,:] = self.co2vec_nlp[self.i2co[i]]
		return self.co2emb, self.co2i, self.i2co, self.ontology_mat

	def BuildModel(self, ngene, nhidden=[1000], use_pretrain=None):
		"""
		Train the model or use the pretrain model
		Parameters
		----------
		train_feature : cell by gene matrix
		train_label: cell by nclass binarize matrix.
		label_emb: embedding vectors of labels (classes)
		genes: gene names of each column in the train feature
		Returns
		-------
		"""

		self.ngene = ngene
		self.use_pretrain = use_pretrain
		#self.label_emb = label_emb
		self.nhidden = nhidden
		if use_pretrain is not None:
			npzfile = np.load(use_pretrain+'.npz',allow_pickle=True)
			self.co2i = npzfile['co2i'].item()
			self.i2co = npzfile['i2co'].item()
			self.genes = npzfile['genes']
			self.co2emb = npzfile['co2emb']
			self.ngene = len(self.genes)
			self.ontology_mat = npzfile['ontology_mat']
			self.nco = npzfile['nco']
			self.nseen = npzfile['nseen']
			self.co2vec_nlp_mat = npzfile['co2vec_nlp_mat']
			self.nhidden = npzfile['nhidden']
			self.ontology_dict = npzfile['ontology_dict'].item()
			self.train_feature_mean = npzfile['train_feature_mean']
		self.model = BilinearNN(self.co2emb, self.nseen, self.ngene, use_pretrain = use_pretrain, nhidden=self.nhidden)
		return self.model

	def ProcessTrainFeature(self, train_feature, train_label, train_genes, test_feature = None, test_genes = None, batch_correct = True, log_transform = True):
		if log_transform is False and np.max(train_feature) > 1000:
			warnings.warn("Max expression is"+str(np.max(train_feature))+'. Consider setting log transform = True\n')
		self.genes = train_genes
		if batch_correct and test_feature is not None and test_genes is not None:
			train_feature, test_feature, selected_train_genes = process_expression(train_feature, test_feature, train_genes, test_genes)
			self.genes = selected_train_genes
		elif log_transform:
			train_feature = np.log1p(train_feature)
			if test_feature is not None:
				test_feature = 	np.log1p(test_feature)
		self.train_feature = train_feature
		self.train_label = train_label
		if test_feature is not None:
			return train_feature, test_feature, self.genes, self.genes
		else:
			return train_feature, self.genes

	def ProcessTestFeature(self, test_feature, test_genes, use_pretrain = None, batch_correct = False, log_transform = True):
		if log_transform is False and np.max(test_feature) > 1000:
			warnings.warn("Max expression is"+str(np.max(test_feature))+'. Consider setting log transform = True\n')
		test_feature = map_genes(test_feature, test_genes, self.genes)
		if use_pretrain is not None:
			if log_transform:
				test_feature = np.log1p(test_feature)
			if batch_correct:
				test_feature = mean_normalization(self.train_feature_mean, test_feature)
		return test_feature

	def Train(self, train_feature, train_label, save_model = None, max_iter=50, minibatch_size = 128, batch_correct = True):
		"""
		Train the model or use the pretrain model
		Parameters
		----------
		train_feature : cell by gene matrix
		train_label: cell by nclass binarize matrix.
		label_emb: embedding vectors of labels (classes)
		Returns
		-------
		"""
		if self.use_pretrain:
			print ('Use pretrained model: ',self.use_pretrain)
			return
		unseen_l = [self.co2i[tp] for tp in self.unseen_co]
		train_label = [self.co2i[tp] for tp in train_label]

		self.train_feature_mean = np.mean(train_feature, axis = 0)
		self.model.read_training_data(train_feature , train_label)
		self.model.optimize(max_iter = max_iter, minibatch_size = minibatch_size, save_model =  save_model)

		if save_model is not None:
			save_model_file = save_model + '.npz'
			np.savez(save_model_file, train_feature_mean = self.train_feature_mean, co2i = self.co2i, co2emb = self.co2emb, nhidden = self.nhidden, i2co = self.i2co, genes = self.genes, nco = self.nco, nseen = self.nseen,
			 ontology_mat = self.ontology_mat, co2vec_nlp_mat = self.co2vec_nlp_mat, ontology_dict = self.ontology_dict)

	def Predict(self, test_feature, test_genes=None, use_normalize=False, refine = True, use_unseen_distance = 2, batch_correct = False):
		"""
		Predict the label for new cells
		"""


		if test_genes is not None:
			test_feature = map_genes(test_feature, test_genes, self.genes)
		else:
			assert(np.shape(test_feature)[1] == self.ngene)
		test_Y_pred_seen = self.model.predict(test_feature)
		test_Y_pred_all = None
		if refine:
			ratio = (self.nco*1./self.nseen)**2
			network = create_propagate_networks_using_nlp(self.co2i, self.ontology_dict, self.ontology_mat, self.co2vec_nlp_mat)
			test_Y_pred_all = extend_prediction_2unseen(test_Y_pred_seen, network, self.nseen, ratio = ratio, use_normalize = use_normalize)
		if use_unseen_distance>=0:
			test_Y_pred_all = create_unseen_candidates(self.cell_type_network_file, self.co2i, self.i2co, self.nseen, use_unseen_distance, test_Y_pred_all)
		return test_Y_pred_seen, test_Y_pred_all, np.argmax(test_Y_pred_all,axis=1)



from setuptools import setup, find_packages

setup(
    name = "OnClass",
    version = "0.97",
    keywords = ("pip", "single cell", "OnClass"),
    description = "Single Cell Annotation based on the Cell Ontology",
    long_description = "Unifying single-cell annotations based on the Cell Ontology",
    license = "MIT Licence",

    url = "https://github.com/wangshenguiuc/OnClass",
    author = "Sheng Wang",
    author_email = "swang91@uw.edu",

    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires=[
		'sentence-transformers',
		'anndata>=0.6.4',
        'fbpca>=1.0',
        'umap-learn>=0.3.10',
        'matplotlib>=2.0.2',
        'numpy>=1.16.4',
        'scipy>=1.3.1',
        'scikit-learn>=0.21.3',
		'scanorama>=1.7.1'
    ]
)
